/*
Physics Toolset
Robert L. Watson
2015

Purpose: Perform various computations with respect to Stencyl's implementation of Box2D
*/

using StringTools;

import com.stencyl.Engine;

import com.stencyl.behavior.Script;
import com.stencyl.behavior.Script.*;
import com.stencyl.behavior.ActorScript;
import com.stencyl.behavior.SceneScript;
import com.stencyl.behavior.TimedTask;

import haxe.ds.GenericStack;
import haxe.ds.*;
import box2D.common.math.B2Vec2;
import box2D.dynamics.B2Fixture;
import com.stencyl.models.Actor;

class PhysicsTools
{
	/*
		[internal]
		This function computes the position (x or y) after n steps.
	*/
	public static function trajectoryPosition( step:Int, startPos:Float, initVel:Float, initGrav:Float ):Float
	{
		var stepConst:Float = getStepSize( ) / 1000.0;
		var vel:Float = initVel * 10 * stepConst;
		var g:Float = initGrav * 10 * stepConst * stepConst;

		return startPos + step * vel + 0.5 * ( step * step + step ) * g;
	}

	/*
		[BLOCK]
		This function computes the position (x or y) after n seconds.
	*/
//	public static function trajectoryPosition3( sec:Float, startPos:Float, initVel:Float, initGrav:Float ):Float
	public static function trajectoryPosition3( sec:Float, startPos:Float, initVel:Float, mode:Int ):Float
	{
		var step:Int = Std.int( sec * 1000 / getStepSize( ) );

		var stepConst:Float = getStepSize( ) / 1000.0;
		var vel:Float = initVel * 10 * stepConst;
//		var g:Float = initGrav * 10 * stepConst * stepConst;
		var g:Float = 0;

		if ( mode == 0 )
			g = getGravity( ).x * 10 * stepConst * stepConst;
		else if ( mode == 1 )
			g = getGravity( ).y * 10 * stepConst * stepConst;

		return startPos + step * vel + 0.5 * ( step * step + step ) * g;
	}

	/*
		[BLOCK]
		This function computes the position of an actor (x or y) after n seconds.
	*/
	public static function trajectoryPosition2( a:Actor, sec:Float, mode:Int ):Float
	{
		var step:Int = Std.int( sec * 1000 / getStepSize( ) );

		if ( mode == 0 )
			return trajectoryPosition( step, a.getX( ), a.getXVelocity( ), getGravity( ).x );
		else if ( mode == 1 )
			return trajectoryPosition( step, a.getY( ), a.getYVelocity( ), getGravity( ).y );
		else if ( mode == 2 )
			return trajectoryPosition( step, a.getX( ) - getScreenX( ), a.getXVelocity( ), getGravity( ).x );
		else if ( mode == 3 )
			return trajectoryPosition( step, a.getY( ) - getScreenY( ), a.getYVelocity( ), getGravity( ).y );
		else
			return -1;
	}


	/*
		[BLOCK]
		Velocity to launch a projectile toward a specified target
		Launch velocity from x1: %1 y1: %2 to x2: %3 y2: %4 at angle %5

		Returns -1 if failed.
	*/
	public static function launchVelocity( x1:Float, y1:Float, x2:Float, y2:Float, angle:Float, mode:Int ):Float
	{
		var n:Float;
		var vx:Float;
		var hDist:Float;

		// 1. Figure out the needed horizontal velocity.
		var arcHeight:Float = Math.abs( x2 - x1 ) * Math.tan( -1 * angle * Math.PI / 180 );

		var stepConst:Float = getStepSize( ) / 1000.0;
		var gy:Float = getGravity( ).y * 10 * stepConst * stepConst;
		var gx:Float = getGravity( ).x * 10 * stepConst * stepConst;

		// 2. Quadratic Formula
		var a:Float = 0.5 / gy;
		var b:Float = 0.5;
		var c:Float = arcHeight;

		var soln1:Float = ( -1 * b + Math.sqrt( b * b - 4 * a * c ) ) / ( 2 * a );
		var soln2:Float = ( -1 * b - Math.sqrt( b * b - 4 * a * c ) ) / ( 2 * a );

		var vy:Float = soln1;

		if ( vy > 0 )
			vy = soln2;

		if ( mode == 1 )
			return vy / ( 10 * stepConst );

trace( "angle %f", angle );
trace( "th %f", arcHeight );
trace( "vy %f", vy);

		// 3. Find number of steps to reach target's height
		a = gy * 0.5;
		b = vy + gy * 0.5;
		c = y1 - y2;

		// 4. Check the discriminant. If it is negative, the likely culprit is that
		//    the launch angle is insufficient to reach the target height.
		//    (arcHeight < y1 - y2)
		if ( b * b - 4 * a * c < 0 )
		{
			trace( "PhysicsTools.launchVelocity: Negative discriminant." );
			return 0;
		}

		soln1 = ( -1 * b + Math.sqrt( b * b - 4 * a * c ) ) / ( 2 * a );
		soln2 = ( -1 * b - Math.sqrt( b * b - 4 * a * c ) ) / ( 2 * a );

trace( "a %f", a);
trace( "b %f", b);
trace( "c %f", c);


trace( "soln1 %f", soln1);
trace( "soln2 %f", soln2);

		n = soln1;

		if ( soln2 > soln1 )
			n = soln2;

		// 4. Find horizontal velocity.
//		hDist = Math.abs( x2 - x1 );
		hDist = x2 - x1;

		vx = ( 1. / n ) * ( hDist - ( 1 / 2 ) * ( n * n + n ) * gx );

trace( "n %f", n);
trace( "x1 %f", x1);
trace( "x2 %f", x2);
trace( "vx %f", vx);

		// 5. Done!
//		if ( x2 < x1 )
//			return -1 * vx / ( 10 * stepConst );
//		else
			return vx / ( 10 * stepConst );
	}

	public static function combinedLaunchVelocity( xVel:Float, yVel:Float ):Float
	{
		return Math.sqrt( xVel * xVel + yVel * yVel );
	}

	public static function combinedLaunchAngle( xVel:Float, yVel:Float, mode:Int ):Float
	{
		if ( mode == 0 )
			return Math.atan2( yVel, xVel ) * 180 / Math.PI;
		else
			return Math.atan2( yVel, xVel );
	}

	/*
		[BLOCK]
		Actor is launched toward a specified target
		Launch %0 to x: %1 y: %2 at angle %3
	*/
	public static function doLaunch( a:Actor, x2:Float, y2:Float, angle:Float )
	{
		var x1:Float = a.getX( );
		var y1:Float = a.getY( );

		var vx:Float = launchVelocity( x1, y1, x2, y2, angle, 0 );
		var vy:Float = launchVelocity( x1, y1, x2, y2, angle, 1 );

//		a.setXVelocity( vx );
//		a.setYVelocity( vy );

		var force:Float = combinedLaunchVelocity( vx, vy );
		var vangle:Float = combinedLaunchAngle( vx, vy, 0 );

		if ( Math.abs( x2 - x1 ) * Math.tan( -1 * angle * Math.PI / 180 ) > y2 - y1 )
		{
			trace( "PhysicsTools.doLaunch: Angle insufficient to reach target height." );
			return;
		}

		a.applyImpulseInDirection( vangle, force );
	}

	public static function angleCheck( x1:Float, y1:Float, x2:Float, y2:Float, angle:Float ):Bool
	{
		if ( Math.abs( x2 - x1 ) * Math.tan( -1 * angle * Math.PI / 180 ) > y2 - y1 )
			return false;
		else
			return true;
	}

	public static function jumpPixel( a:Actor, x:Float, y:Float, dir:Int )
	{
		var angle:Float = Math.atan( y / x ) * 180 / Math.PI;

		doLaunch( a, a.getX( ) + ( dir * x ), a.getY( ), angle );
	}

	public static function jumpTile( a:Actor, x:Float, y:Float, dir:Int )
	{
		var angle:Float = Math.atan( y / x ) * 180 / Math.PI;

		doLaunch( a, a.getX( ) + ( dir * x * getTileWidth( ) ), a.getY( ), angle );
	}

	/*
		[BLOCK]
		This function implements a linear spline-based arc raycasting check.
		It returns the first actor an arc-based trajectory will intersect
	*/
//	public static function arcRayCast( x0:Float, y0:Float, startsec:Float, sec:Float, steps:Float, xvel:Float, yvel:Float ):Actor
	public static function arcRayCast( x0:Float, y0:Float, startsec:Float, sec:Float, xvel:Float, yvel:Float ):Actor
	{
		var vx:Float = xvel;
		var vy:Float = yvel;
		var gx:Float = getGravity( ).x;
		var gy:Float = getGravity( ).y;

		var steps = 0.1;

		var stepMin:Int = Std.int( startsec * 1000 / getStepSize( ) );
		var stepMax:Int = Std.int( sec * 1000 / getStepSize( ) );
		var incr:Int = Std.int( steps * 1000 / getStepSize( ) );

		var engine = Engine.engine;
		var world = engine.world;

		var i:Int = stepMin;

//		var x1:Float = x0;
//		var y1:Float = y0;

		var x1:Float = trajectoryPosition( i, x0, vx, gx );
		var y1:Float = trajectoryPosition( i, y0, vy, gy );

		while ( i < stepMax )
		{
			var x2:Float = trajectoryPosition( i, x0, vx, gx );
			var y2:Float = trajectoryPosition( i, y0, vy, gy );

			var p1:B2Vec2 = new B2Vec2( Engine.toPhysicalUnits( x1 ), Engine.toPhysicalUnits( y1 ) );
			var p2:B2Vec2 = new B2Vec2( Engine.toPhysicalUnits( x2 ), Engine.toPhysicalUnits( y2 ) );

			var result:B2Fixture = world.rayCastOne( p1, p2 );

			if( result != null )
			{
				return result.getUserData( );
			}
			else
			{
				x1 = x2;
				y1 = y2;
			}

			i += incr;
		}

		return null;
	}

	/*
		[BLOCK]
		This function computes the velocity (x or y) to reach a given position after n seconds.
	*/
/*	public static function predictedVelocity( targetPos:Float, startPos:Float, sec:Float, initGrav:Float ):Float
	{
		var step:Int = Std.int( sec * 1000 / getStepSize( ) );

		var stepConst:Float = getStepSize( ) / 1000.0;
		var g:Float = initGrav * 10 * stepConst * stepConst;

		vel = ( targetPos - startPos - 0.5 * ( step * step + step ) * g ) * 1. / step;

		return vel / ( 10 * stepConst );
	} */

}


